@@ -6,12 +6,13 @@
 
 __author__ = """luphord"""
 __email__ = """luphord@protonmail.com"""
-__version__ = """0.3.0"""
+__version__ = """0.4.0"""
 
 
 from argparse import ArgumentParser, Namespace
 from abc import ABC, abstractmethod
-from typing import Final, Union, Mapping, Tuple, Set
+from itertools import groupby
+from typing import Final, Union, Mapping, Tuple, Set, Iterable
 from numbers import Real
 import numpy as np
 from dataclasses import dataclass
@@ -22,6 +23,28 @@ _null_ccy = "NNN"
 ArrayLike = Union[np.ndarray, float]
 
 
+@dataclass
+class SimpleCashflows:
+    cashflows: np.ndarray
+    currencies: np.ndarray
+    dates: np.ndarray
+    nsim: int
+
+    def __init__(
+        self, cashflows: np.ndarray, currencies: np.ndarray, dates: np.ndarray
+    ):
+        assert cashflows.dtype == np.float64
+        assert cashflows.ndim == 2, f"Array must have ndim 2, got {cashflows.ndim}"
+        assert currencies.dtype == (np.unicode_, _ccy_letters)
+        assert currencies.shape == (cashflows.shape[1],)
+        assert dates.dtype == "datetime64[D]"
+        assert dates.shape == (cashflows.shape[1],)
+        self.cashflows = cashflows
+        self.currencies = currencies
+        self.dates = dates
+        self.nsim = cashflows.shape[0]
+
+
 class SimulatedCashflows:
     dtype: Final = np.dtype([("date", "datetime64[D]"), ("value", np.float64)])
     cashflows: Final[np.ndarray]
@@ -41,6 +64,39 @@ class SimulatedCashflows:
         self.nsim = cashflows.shape[0]
         self.ncashflows = cashflows.shape[1]
 
+    def _split_by_date(self) -> Iterable[Tuple[np.ndarray, str, np.datetime64]]:
+        for i, cf in enumerate(self.cashflows.T):
+            if self.currencies[i] != _null_ccy:
+                for dt in np.unique(cf["date"]):
+                    if not np.isnat(dt):
+                        cf_dt = cf.copy()
+                        cf_dt[cf_dt["date"] != dt] = 0
+                        yield cf_dt, self.currencies[i], dt
+
+    def _group_cashflows(self) -> Iterable[Tuple[np.ndarray, str, np.datetime64]]:
+        def grpkey(
+            entry: Tuple[np.ndarray, str, np.datetime64]
+        ) -> Tuple[np.datetime64, str]:
+            cf, ccy, dt = entry
+            return (dt, ccy)
+
+        for key, entries in groupby(sorted(self._split_by_date(), key=grpkey), grpkey):
+            dt, ccy = key
+            yield sum(cf["value"] for cf, _, _ in entries), ccy, dt
+
+    def to_simple_cashflows(self) -> SimpleCashflows:
+        grouped_cf = list(self._group_cashflows())
+        assert grouped_cf, "Got no cashflows to convert"
+        numcf = len(grouped_cf)
+        cashflows = np.ndarray((self.nsim, numcf), dtype=np.float64)
+        currencies = np.ndarray((numcf,), dtype=(np.unicode_, _ccy_letters))
+        dates = np.ndarray((numcf,), dtype="datetime64[D]")
+        for i, (cf, ccy, dt) in enumerate(grouped_cf):
+            cashflows[:, i] = cf
+            currencies[i] = ccy
+            dates[i] = dt
+        return SimpleCashflows(cashflows, currencies, dates)
+
 
 class IndexedCashflows:
     dtype: Final = np.dtype([("index", np.int), ("value", np.float64)])
@@ -227,6 +283,27 @@ class Model:
     def generate_cashflows(self, contract: "Contract") -> SimulatedCashflows:
         return contract.generate_cashflows(self.eval_date_index, self).apply_index()
 
+    def generate_simple_cashflows(self, contract: "Contract") -> SimpleCashflows:
+        return self.generate_cashflows(contract).to_simple_cashflows()
+
+    def generate_simple_cashflows_in_currency(
+        self, contract: "Contract", currency: str
+    ) -> SimpleCashflows:
+        return (
+            self.in_currency(
+                contract.generate_cashflows(self.eval_date_index, self), currency
+            )
+            .apply_index()
+            .to_simple_cashflows()
+        )
+
+    def generate_simple_cashflows_in_numeraire_currency(
+        self, contract: "Contract"
+    ) -> SimpleCashflows:
+        return self.generate_simple_cashflows_in_currency(
+            contract, self.numeraire_currency
+        )
+
     def in_currency(
         self, cashflows: IndexedCashflows, currency: str
     ) -> IndexedCashflows:
@@ -253,6 +330,31 @@ class Model:
     def in_numeraire_currency(self, cashflows: IndexedCashflows) -> IndexedCashflows:
         return self.in_currency(cashflows, self.numeraire_currency)
 
+    def discount(self, cashflows: IndexedCashflows) -> np.ndarray:
+        cfnum = self.in_numeraire_currency(cashflows)
+        discounted = np.zeros(cfnum.cashflows.shape, dtype=np.float64)
+        num_0 = self.numeraire[:, 0]
+        for i, cf in enumerate(cfnum.cashflows.T):
+            num_t = DateIndex(cf["index"]).index_column(self.numeraire)
+            num_t[cf["index"] < 0] = 1  # cashflow happens never anyway
+            discounted[:, i] = cf["value"] / num_t * num_0
+        return discounted
+
+    def evaluate(
+        self, cashflows_or_contract: Union[IndexedCashflows, "Contract"]
+    ) -> float:
+        cf: IndexedCashflows
+        if isinstance(cashflows_or_contract, IndexedCashflows):
+            cf = cashflows_or_contract
+        elif isinstance(cashflows_or_contract, Contract):
+            cf = cashflows_or_contract.generate_cashflows(self.eval_date_index, self)
+        else:
+            raise TypeError(
+                "Expecting IndexedCashflows or Contract, "
+                f"got {type(cashflows_or_contract)}"
+            )
+        return self.discount(cf).sum(axis=1).mean(axis=0)
+
 
 class ObservableFloat(ABC):
     """Abstract base class for all observables of underlying type float,
@@ -294,6 +396,9 @@ class Stock(ObservableFloat):
     def simulate(self, model: Model) -> np.ndarray:
         return model.simulated_stocks[self.identifier]
 
+    def __str__(self) -> str:
+        return f"Stock({self.identifier})"
+
 
 @dataclass
 class FX(ObservableFloat):
@@ -306,6 +411,9 @@ class FX(ObservableFloat):
     def simulate(self, model: Model) -> np.ndarray:
         return model.get_simulated_fx(self.base_currency, self.counter_currency)
 
+    def __str__(self) -> str:
+        return f"FX({self.base_currency}/{self.counter_currency})"
+
 
 @dataclass
 class KonstFloat(ObservableFloat):
@@ -316,6 +424,9 @@ class KonstFloat(ObservableFloat):
     def simulate(self, model: Model) -> np.ndarray:
         return self.constant * np.ones(model.shape, dtype=np.float)
 
+    def __str__(self) -> str:
+        return str(self.constant)
+
 
 class ObservableBool(ABC):
     """Abstract base class for all observables of underlying type bool"""
@@ -343,6 +454,9 @@ class Not(ObservableBool):
     def simulate(self, model: Model) -> np.ndarray:
         return ~self.observable.simulate(model)
 
+    def __str__(self) -> str:
+        return f"~({self.observable})"
+
 
 @dataclass
 class AndObservable(ObservableBool):
@@ -354,6 +468,9 @@ class AndObservable(ObservableBool):
     def simulate(self, model: Model) -> np.ndarray:
         return self.observable1.simulate(model) & self.observable2.simulate(model)
 
+    def __str__(self) -> str:
+        return f"({self.observable1}) & ({self.observable2})"
+
 
 @dataclass
 class OrObservable(ObservableBool):
@@ -365,6 +482,9 @@ class OrObservable(ObservableBool):
     def simulate(self, model: Model) -> np.ndarray:
         return self.observable1.simulate(model) | self.observable2.simulate(model)
 
+    def __str__(self) -> str:
+        return f"({self.observable1}) | ({self.observable2})"
+
 
 @dataclass
 class GreaterOrEqualThan(ObservableBool):
@@ -376,6 +496,9 @@ class GreaterOrEqualThan(ObservableBool):
     def simulate(self, model: Model) -> np.ndarray:
         return self.observable1.simulate(model) >= self.observable2.simulate(model)
 
+    def __str__(self) -> str:
+        return f"{self.observable1} >= {self.observable2}"
+
 
 @dataclass
 class GreaterThan(ObservableBool):
@@ -387,6 +510,9 @@ class GreaterThan(ObservableBool):
     def simulate(self, model: Model) -> np.ndarray:
         return self.observable1.simulate(model) > self.observable2.simulate(model)
 
+    def __str__(self) -> str:
+        return f"{self.observable1} > {self.observable2}"
+
 
 @dataclass
 class At(ObservableBool):
@@ -399,6 +525,9 @@ class At(ObservableBool):
         assert mask.any(), f"{self.date} not contained in dategrid"
         return np.repeat(mask, model.nsim, axis=0)
 
+    def __str__(self) -> str:
+        return str(self.date)
+
 
 class Contract(ABC):
     """Abstract base class for all contracts"""
@@ -423,6 +552,9 @@ class Zero(Contract):
         ccys = np.array([_null_ccy], dtype=(np.unicode_, _ccy_letters))
         return IndexedCashflows(cf, ccys, model.dategrid)
 
+    def __str__(self) -> str:
+        return "Zero"
+
 
 @dataclass
 class One(Contract):
@@ -441,6 +573,9 @@ class One(Contract):
         ccys = np.array([self.currency], dtype=(np.unicode_, _ccy_letters))
         return IndexedCashflows(cf, ccys, model.dategrid)
 
+    def __str__(self) -> str:
+        return f"One({self.currency})"
+
 
 @dataclass
 class Give(Contract):
@@ -454,6 +589,9 @@ class Give(Contract):
     ) -> IndexedCashflows:
         return -self.contract.generate_cashflows(acquisition_idx, model)
 
+    def __str__(self) -> str:
+        return f"Give({self.contract})"
+
 
 @dataclass
 class And(Contract):
@@ -469,6 +607,9 @@ class And(Contract):
         cf2 = self.contract2.generate_cashflows(acquisition_idx, model)
         return cf1 + cf2
 
+    def __str__(self) -> str:
+        return f"And({self.contract1}, {self.contract2})"
+
 
 @dataclass
 class Or(Contract):
@@ -501,6 +642,9 @@ class Or(Contract):
         cf2.cashflows["value"][choose1] = 0
         return cf1 + cf2
 
+    def __str__(self) -> str:
+        return f"Or({self.contract1}, {self.contract2})"
+
 
 @dataclass
 class Cond(Contract):
@@ -525,6 +669,9 @@ class Cond(Contract):
         cf2.cashflows["value"][never, :] = 0
         return cf1 + cf2
 
+    def __str__(self) -> str:
+        return f"Cond({self.observable}, {self.contract1}, {self.contract2})"
+
 
 @dataclass
 class Scale(Contract):
@@ -545,6 +692,9 @@ class Scale(Contract):
         assert obs.shape[0] == model.nsim
         return cf * obs
 
+    def __str__(self) -> str:
+        return f"Scale({self.observable}, {self.contract})"
+
 
 @dataclass
 class When(Contract):
@@ -560,6 +710,9 @@ class When(Contract):
         idx = acquisition_idx.next_after(self.observable.simulate(model))
         return self.contract.generate_cashflows(idx, model)
 
+    def __str__(self) -> str:
+        return f"When({self.observable}, {self.contract})"
+
 
 @dataclass
 class Anytime(Contract):
@@ -574,6 +727,9 @@ class Anytime(Contract):
     ) -> IndexedCashflows:
         raise NotImplementedError()
 
+    def __str__(self) -> str:
+        return f"Anytime({self.observable}, {self.contract})"
+
 
 @dataclass
 class Until(Contract):
@@ -590,6 +746,9 @@ class Until(Contract):
         ko_idx = acquisition_idx.next_after(self.observable.simulate(model))
         return cf.zero_after(ko_idx)
 
+    def __str__(self) -> str:
+        return f"Until({self.observable}, {self.contract})"
+
 
 class ResolvableContract(Contract):
     @abstractmethod
@@ -601,6 +760,9 @@ class ResolvableContract(Contract):
     ) -> IndexedCashflows:
         return self.resolve().generate_cashflows(acquisition_idx, model)
 
+    def __str__(self) -> str:
+        return str(self.resolve())
+
 
 @dataclass
 class ZeroCouponBond(ResolvableContract):
@@ -623,7 +785,44 @@ class EuropeanOption(ResolvableContract):
         return When(At(self.maturity), Or(self.contract, Zero()))
 
 
-class BrownianMotion:
+class StochasticProcess(ABC):
+    """Base class for stochastic processes."""
+
+    @abstractmethod
+    def simulate(self, t: np.array, n: int, rnd: np.random.RandomState) -> np.array:
+        pass
+
+    @abstractmethod
+    def expected(self, t: np.ndarray) -> np.ndarray:
+        pass
+
+    @abstractmethod
+    def stddev(self, t: np.ndarray) -> np.ndarray:
+        pass
+
+    def moment_match(self, t: np.ndarray, paths: np.ndarray) -> np.ndarray:
+        assert t.ndim == 1, "One dimensional time vector required"
+        ndates = t.size
+        assert ndates > 0, "At least one time point is required"
+        assert paths.ndim == 2
+        assert paths.shape[1] == ndates
+        target_mean = self.expected(t)
+        target_stddev = self.stddev(t)
+        actual_mean = paths.mean(axis=0)
+        actual_stddev = paths.std(axis=0)
+        assert (
+            target_stddev[actual_stddev == 0] == 0
+        ), "Cannot scale actual standard deviation of zero to any other value than zero"
+        actual_stddev[actual_stddev == 0] = 1  # prevent division by zero
+        return (paths - actual_mean) / actual_stddev * target_stddev + target_mean
+
+    def simulate_with_moment_matching(
+        self, t: np.array, n: int, rnd: np.random.RandomState
+    ) -> np.array:
+        return self.moment_match(t, self.simulate(t, n, rnd))
+
+
+class BrownianMotion(StochasticProcess):
     """Brownian Motion (Wiener Process) with optional drift."""
 
     def __init__(self, mu: float = 0.0, sigma: float = 1.0):
@@ -640,8 +839,14 @@ class BrownianMotion:
         W_drift = W * np.sqrt(dt) * self.sigma + self.mu * dt
         return np.cumsum(W_drift, axis=1)
 
+    def expected(self, t: np.ndarray) -> np.ndarray:
+        return self.mu * t
 
-class GeometricBrownianMotion:
+    def stddev(self, t: np.ndarray) -> np.ndarray:
+        return np.sqrt(self.sigma ** 2 * t)
+
+
+class GeometricBrownianMotion(StochasticProcess):
     """Geometric Brownian Motion.(with optional drift)."""
 
     def __init__(self, mu: float = 0.0, sigma: float = 1.0):
@@ -658,6 +863,41 @@ class GeometricBrownianMotion:
         W = np.cumsum(dW, axis=0)
         return np.exp(self.sigma * W.T + (self.mu - self.sigma ** 2 / 2) * t)
 
+    def expected(self, t: np.ndarray) -> np.ndarray:
+        return np.exp(self.mu * t)
+
+    def stddev(self, t: np.ndarray) -> np.ndarray:
+        return np.sqrt(np.exp(2 * self.mu * t) * (np.exp(self.sigma ** 2 * t) - 1))
+
+
+def _get_year_fractions(dategrid: np.ndarray) -> np.ndarray:
+    assert dategrid.dtype == "datetime64[D]"
+    return (dategrid - dategrid[0]).astype(np.float64) / 365
+
+
+def simulate_equity_black_scholes_model(
+    stock: str,
+    currency: str,
+    S0: float,
+    dategrid: np.array,
+    sigma: float,
+    r: float,
+    n: int,
+    rnd: np.random.RandomState,
+    use_moment_matching: bool = False,
+) -> Model:
+    assert dategrid.dtype == "datetime64[D]"
+    ndates = dategrid.size
+    yearfractions = _get_year_fractions(dategrid)
+    gbm = GeometricBrownianMotion(r, sigma)
+    s = S0 * (
+        gbm.simulate_with_moment_matching(yearfractions, n, rnd)
+        if use_moment_matching
+        else gbm.simulate(yearfractions, n, rnd)
+    )
+    numeraire = np.repeat(np.exp(r * yearfractions).reshape((1, ndates)), n, axis=0)
+    return Model(dategrid, {}, {stock: s}, numeraire, currency)
+
 
 parser = ArgumentParser(description=__doc__)
 parser.add_argument(
